#include "stdint.h"
#include "s3c44b0x.h"
#include "font.h"
#include "uart.h"

#define RGB(r,g,b) ((r & 0xE0) | ((g & 0xE0) >> 3) | (b >> 6))
#define SCREEN_WIDTH   240
#define SCREEN_HEIGHT  160
#define FIREMAX        100
#define RAND_MAX 2147483647

#define NUMBER_OF_STARS 1020

typedef struct {
    float xpos, ypos;
    short zpos, speed;
    unsigned char color;
} STAR;

static STAR stars[NUMBER_OF_STARS];
static unsigned char fire[SCREEN_WIDTH * SCREEN_HEIGHT];
static unsigned char colors[256];
uint32_t button_state = 0x00;
const uint8_t color_lookup[] = { 0xff,  /* white */
    0xe0,                       /* red */
    0x1c,                       /* green */
    0x03,                       /* blue */
    0xe0 | 0x1c,                /* red + green */
    0x1c | 0x03,                /* green + blue */
    0xe0 | 0x03,                /* red + blue */
    0,                          /* black */
};


void button_init()
{
    *pupg &= ~(0x00 << 0);      /* 0=pullup */
    *pcong &= ~(0x03 << 0);     /* mask out 2 bits for pin control */
    *pcong |= (0x00 << 0);      /* 0=input (1=output) */

    *pupg &= ~(0x00 << 1);      /* 0=pullup */
    *pcong &= ~(0x03 << 2);     /* mask out 2 bits for pin control */
    *pcong |= (0x00 << 2);      /* 0=input (1=output) */

    *pupg &= ~(0x00 << 2);      /* 0=pullup */
    *pcong &= ~(0x03 << 4);     /* mask out 2 bits for pin control */
    *pcong |= (0x00 << 4);      /* 0=input (1=output) */

    *pupg &= ~(0x00 << 3);      /* 0=pullup */
    *pcong &= ~(0x03 << 6);     /* mask out 2 bits for pin control */
    *pcong |= (0x00 << 6);      /* 0=input (1=output) */

    *pupg &= ~(0x00 << 4);      /* 0=pullup */
    *pcong &= ~(0x03 << 8);     /* mask out 2 bits for pin control */
    *pcong |= (0x00 << 8);      /* 0=input (1=output) */
}

uint32_t read_button()
{
    uint32_t ip, nb;
    ip = *pdatg;
    ip = (~ip) & 0x1f;
    if(ip !=  button_state) {
        uart_tx_string((const uint8_t*)"\r\n  in read_button  ");
        uart_tx_string((const uint8_t*)itoh(ip));
        uart_tx_string((const uint8_t*)"\r\n");
 
    }
    if (ip == button_state)
        return (0);
    if (ip > button_state) {
        nb = ip - button_state;
        button_state = ip;
        return (nb);
    }
    button_state = ip;
    return (0);
}


void delay_ms(uint32_t delay)
{
    /* Delay hand tuned for use with 66.428MHz clock
       (PLLCON = 0x55051, PLL reference of 10MHz).
       Assumes code is running from cached memory.
       fixme: use a hardware timer instead.
     */

    uint32_t ms;
    uint32_t i;
    for (ms = 0; ms < delay; ms++)
        for (i = 0; i < 13288; i++);
}

void cpu_init()
{
    /* Set clock rate to 66.428MHz, just in case the JB has
       it set wrong
     */
    *pllcon = 0x55051;

    /* Be sure we're in cacheable memory
       Can the framebuffer be in cache too?
     */
    *ncachbe0 = 0x80002000;
    *ncachbe1 = 0x0;
}

uint32_t *framebuffer_start()
{
    /* Return with starting address of LCD framebuffer */
    return (uint32_t *) (((*lcdsaddr1) & 0x07ffffff) << 1);
}

void clear_lcd()
{
    uint32_t i;
    uint32_t *addr = framebuffer_start();

    /* fill entire framebuffer, 32 bits at a whack */
    for (i = 0; i < 240 * 160; i += 4) {
        *addr = 0;
        ++addr;
    }
}

void set_pixel(uint32_t x, uint32_t y, uint8_t color)
{
    uint8_t *addr = (uint8_t *) framebuffer_start();
    addr += x;
    addr += (240 * y);
    *addr = color;
}

void lcd_init()
{
    /* Set ENVID (bit 0 of lcdcon1) to 0 to disable panel */
    *lcdcon1 = 0x10C0DF40;      /* LCDCON1  (LCD Control 1) */

    /* Configure the LCD panel, using settings borrowed from
       the built in self test code.
       Looks like it's one byte/pixel, using the palette LUT registers.
     */
    *lcdcon2 = 0x0141649F;      /* LCDCON2  (LCD Control 2) */
    *lcdcon3 = 0x00000000;      /* LCDCON3  (LCD Control 3) */
    *lcdsaddr1 = 0x1E0F288A;    /* LCDSADDR1  (Frame Upper Buffer Start Address 1) */
    *lcdsaddr2 = 0x21AF738A;    /* LCDSADDR2  (Frame Lower Buffer Start Address 2) */
    *lcdsaddr3 = 0x00000078;    /* LCDSADDR3  (Virtual Screen Address) */
    *redlut = 0xFDB96420;       /* REDLUT  (RED Lookup Table) */
    *greenlut = 0xFDB96420;     /* GREENLUT  (GREEN Lookup Table) */
    *bluelut = 0x0000FB40;      /* BLUELUT  (BLUE Lookup Table) */
    *dp1_2 = 0x0000A5A5;        /* DP1_2  (Dithering Pattern duty 1/2) */
    *dp4_7 = 0x0BA5DA65;        /* DP4_7  (Dithering Pattern duty 4/7) */
    *dp3_5 = 0x000A5A5F;        /* DP3_5  (Dithering Pattern duty 3/5) */
    *dp2_3 = 0x00000D6B;        /* DP2_3  (Dithering Pattern duty 2/3) */
    *dp5_7 = 0x0EB7B5ED;        /* DP5_7  (Dithering Pattern duty 5/7) */
    *dp3_4 = 0x00007DBE;        /* DP3_4  (Dithering Pattern duty 3/4) */
    *dp4_5 = 0x0007EBDF;        /* DP4_5  (Dithering Pattern duty 4/5) */
    *dp6_7 = 0x07FDFBFE;        /* DP6_7  (Dithering Pattern duty 6/7) */
    *dithmode = 0x00000000;     /* DITHMODE  (Dithering Mode) */

    /* Enable the panel now that the settings are in place */
    *lcdcon1 = 0x10C0DF41;      /* LCDCON1  (LCD Control 1) */
}

void draw_char(uint32_t x, uint32_t y, uint8_t c, uint8_t color)
{
    /* Draw a single ascii character on screen at location x,y */
    uint32_t i;
    uint32_t j;

    /* Chars are 7 pixels tall, 5 pixels wide */
    for (i = 0; i < 7; i++) {
        for (j = 0; j < 5; j++) {
            if (((ascii_lookup[c - 32][i]) << (3 + j)) & 0x80) {
                set_pixel(x + j, y + i, color);
            }
        }
    }
}

void draw_string(const uint8_t * s, uint32_t x, uint32_t y, uint8_t color)
{
    /* Draw a string of ascii chars on screen */
    while (*s) {
        draw_char(x, y, *s, color);
        x += 6;
        s++;
    }
}

uint32_t rand()
{
    /* This is a really bad linear congruential generator */
    static uint32_t r;
    r = (1664525L * r) + 1013904223L;
    return r;
}

void fire_init()
{
    int i;

    for (i = 0; i < 32; ++i) {

        colors[i] = RGB(0, 0, (i << 1));
        colors[i + 32] = RGB((i << 3), 0, (64 - (i << 1)));
        colors[i + 64] = RGB(255, (i << 3), 0);
        colors[i + 96] = RGB(255, 255, (i << 2));
        colors[i + 128] = RGB(255, 255, (64 + (i << 2)));
        colors[i + 160] = RGB(255, 255, (128 + (i << 2)));;
        colors[i + 192] = RGB(255, 255, (192 + i));
        colors[i + 224] = RGB(255, 255, (224 + i));
    }

}


int fire_main()
{
    unsigned char *image;
    unsigned char index;
    int temp;
    int i, j;

    fire_init();
    uart_tx_string("fire init done\r\n");
    clear_lcd();
    draw_string("Juicebox Fire by OPTIXX ", 10, 10, 0xff);

    while (1) {
        if(read_button())
            break;
 
        j = SCREEN_WIDTH * (SCREEN_HEIGHT - 1);
        for (i = 0; i < SCREEN_WIDTH - 1; i++) {
            int random = 1 + (int) (12.0 * (rand() / (RAND_MAX + 1.0)));
            if (random > 9)     /* the lower the value, the intenser the fire, compensate a lower value with a higher decay value */
                fire[j + i] = 255;      /*maximum heat */
            else
                fire[j + i] = 0;
        }

        /* move fire upwards, start at bottom */

        for (index = 0; index < FIREMAX; ++index) {
            for (i = 0; i < SCREEN_WIDTH - 1; ++i) {
                if (i == 0) {   /* at the left border */
                    temp = fire[j];
                    temp += fire[j + 1];
                    temp += fire[j - SCREEN_WIDTH];
                    temp /= 3;
                } else if (i == SCREEN_WIDTH - 1) {     /* at the right border */
                    temp = fire[j + i];
                    temp += fire[j - SCREEN_WIDTH + i];
                    temp += fire[j + i - 1];
                    temp /= 3;
                } else {
                    temp = fire[j + i];
                    temp += fire[j + i + 1];
                    temp += fire[j + i - 1];
                    temp += fire[j - SCREEN_WIDTH + i];
                    temp >>= 2;
                }
                if (temp > 1)
                    temp -= 1;  /* decay */

                fire[j - SCREEN_WIDTH + i] = temp;
            }
            j -= SCREEN_WIDTH;
        }

        image = (uint8_t *) framebuffer_start();
        image = image + (SCREEN_HEIGHT * SCREEN_WIDTH);

        for (i = SCREEN_HEIGHT - 3; i >= SCREEN_HEIGHT - FIREMAX; --i) {
            for (j = SCREEN_WIDTH - 1; j >= 0; --j) {
                *image = colors[fire[i * SCREEN_WIDTH + j]];
                image--;
            }
        }

    }

    return 0;                   /* never reached */
}

void star_init(STAR * star, int i)
{
    /* randomly init stars, generate them around the center of the screen */

    star->xpos = -10.0 + (20.0 * (rand() / (RAND_MAX + 1.0)));
    star->ypos = -10.0 + (20.0 * (rand() / (RAND_MAX + 1.0)));

    star->xpos *= 3072.0;       /*change viewpoint */
    star->ypos *= 3072.0;

    star->zpos = i;
    star->speed = 2 + (int) (2.0 * (rand() / (RAND_MAX + 1.0)));

    star->color = i >> 2;       /*the closer to the viewer the brighter */
}


void starfield_init()
{
    int i;
    colors[0] = 0;

    for (i = 0; i < 255; ++i) {
        colors[255 - i] = RGB(i, i, i);
    }


    for (i = 0; i < NUMBER_OF_STARS; i++) {
        star_init(stars + i, i + 1);
    }

}

int starfield_main()
{
    int centerx, centery;
    int i, tempx, tempy;
    unsigned char *image;
    starfield_init();
    clear_lcd();
    draw_string("Juicebox Starfield by OPTIXX ", 10, 10, 0xff);

    centerx = SCREEN_WIDTH >> 1;
    centery = SCREEN_HEIGHT >> 1;

    while (1) {

        clear_lcd();
        if(read_button())
            break;
        for (i = 0; i < NUMBER_OF_STARS; i++) {
            stars[i].zpos -= stars[i].speed;

            if (stars[i].zpos <= 0) {
                star_init(stars + i, i + 1);
            }

            /*compute 3D position */
            tempx = (stars[i].xpos / stars[i].zpos) + centerx;
            tempy = (stars[i].ypos / stars[i].zpos) + centery;

            if (tempx < 0 || tempx > SCREEN_WIDTH - 1 || tempy < 0 || tempy > SCREEN_HEIGHT - 1) {      /*check if a star leaves the screen */
                star_init(stars + i, i + 1);
                continue;
            }
            image = (uint8_t *) framebuffer_start();

            *((unsigned char *) image + (tempy * SCREEN_WIDTH) + tempx) = stars[i].color;
        }
        delay_ms(10);
    }
    return 0;                   /* never reached */
}



int main(void)
{
    cpu_init();
    lcd_init();
    uart_init();
    uart_tx_string("init done\r\n");
    for (;;) {
        uart_tx_string("starfield_main\r\n");
        starfield_main();
        uart_tx_string("fire_main\r\n");
        fire_main();

    }
    return 0;
}
