#include "stdint.h"
#include "s3c44b0x.h"

void uart_init()
{
    /* Assign TX and RX pins to the UART */

    *pcone &= ~0x0c;
    *pcone |= 0x08;



    /* 8 bits, 2 stop bits, no parity, non-IrDA mode */
    *ULCON0 = 0x03;

    /* Polled RX, polled TX, no break, no loopback, no error interrupts,
       no RX fifo timeouts, TX level interrupts
     */
    *UCON0 = 0x205;

    /* Clear and enable FIFOs */
    *UFCON0 = 7;

    /* Disable FIFOs */
    *UFCON0 = 0;

    /* Disable automatic flow control on handshake lines */
    *UMCON0 = 0;

    /* Set baud rate divisor
       0x23 = 115.2k baud with 66.428MHz clock
       (OK, it's really 115.327kbaud, or 0.11% fast).
     */
    *UBRDIV0 = 0X23;
}

void uart_tx_string(const uint8_t * s)
{
    while (*s) {
        while (!(*UTRSTAT0 & 0x02)) {
        };                      /* spin until TX ready */
        *UTXH0 = *s;
        s++;
    }
}

void uart_tx_char(const uint8_t c)
{
    while (!(*UTRSTAT0 & 0x02)) {
    };                          /* spin until TX ready */
    *UTXH0 = c;
}

const char *itoh(uint32_t x)
{
    /* Nasty hack to convert 32 bit unsigned value
       to a string of 8 hex digits.
       Static allocation, so it's not re-entrant
     */
    uint8_t d;
    uint32_t i;
    static char s[10];
    s[8] = 0;

    for (i = 0; i < 8; i++) {
        d = ((x >> ((7 - i) * 4)) & 0x0f);

        if (d <= 9)
            s[i] = d + '0';
        else
            s[i] = d - 0x0a + 'A';
    }
    return s;
}
